"""Calculate graph edge bearings."""

import math

import numpy as np


def get_bearing(origin_point, destination_point):
    """
    Calculate the bearing between two lat-lng points.

    Each tuple should represent (lat, lng) as decimal degrees.

    Parameters
    ----------
    origin_point : tuple
        (lat, lng)
    destination_point : tuple
        (lat, lng)

    Returns
    -------
    bearing : float
        the compass bearing in decimal degrees from the origin point to the
        destination point
    """
    if not (isinstance(origin_point, tuple) and isinstance(destination_point, tuple)):
        raise TypeError("origin_point and destination_point must be (lat, lng) tuples")

    # get latitudes and the difference in longitude, as radians
    lat1 = math.radians(origin_point[0])
    lat2 = math.radians(destination_point[0])
    diff_lng = math.radians(destination_point[1] - origin_point[1])

    # calculate initial bearing from -180 degrees to +180 degrees
    x = math.sin(diff_lng) * math.cos(lat2)
    y = math.cos(lat1) * math.sin(lat2) - (math.sin(lat1) * math.cos(lat2) * math.cos(diff_lng))
    initial_bearing = math.atan2(x, y)

    # normalize initial bearing to 0-360 degrees to get compass bearing
    initial_bearing = math.degrees(initial_bearing)
    bearing = (initial_bearing + 360) % 360

    return bearing


def add_edge_bearings(G, precision=1):
    """
    Add `bearing` attributes to all graph edges.

    Calculate the compass bearing from origin node to destination node for
    each edge in the directed graph then add each bearing as a new edge
    attribute.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    precision : int
        decimal precision to round bearing

    Returns
    -------
    G : networkx.MultiDiGraph
        graph with edge bearing attributes
    """
    for u, v, data in G.edges(keys=False, data=True):

        if u == v:
            # a self-loop has an undefined compass bearing
            data["bearing"] = np.nan

        else:
            # calculate bearing from edge's origin to its destination
            origin_point = (G.nodes[u]["y"], G.nodes[u]["x"])
            destination_point = (G.nodes[v]["y"], G.nodes[v]["x"])
            bearing = get_bearing(origin_point, destination_point)
            data["bearing"] = round(bearing, precision)

    return G



"""Plot spatial geometries, street networks, and routes."""

import os
import warnings

import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from descartes import PolygonPatch
from matplotlib.collections import LineCollection
from matplotlib.collections import PatchCollection
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

from . import graph
from . import settings
from . import simplification
from . import utils
from . import utils_geo
from . import utils_graph


def _warn_deprecated_params(**kwargs):
    """
    Warn about deprecated params.

    Parameters
    ----------
    kwargs

    Returns
    -------
    None
    """
    params = [k for k, v in kwargs.items() if v is not None]
    if len(params) > 0:

        param_str = ", ".join(params)
        msg = f"The {param_str} parameter(s) have been deprecated and will be removed in the next release. "

        cond1 = "fig_height" in kwargs and kwargs["fig_height"] is not None
        cond2 = "fig_width" in kwargs and kwargs["fig_width"] is not None
        if cond1 or cond2:
            msg += "Note, fig_height and fig_width are replaced by the figsize parameter, use that instead. "

        cond1 = "filename" in kwargs and kwargs["filename"] is not None
        cond2 = "file_format" in kwargs and kwargs["file_format"] is not None
        if cond1 or cond2:
            msg += "Note, filename and file_format are replaced by the filepath parameter, use that instead. "

        if "orig_dest_node_size" in kwargs and kwargs["orig_dest_node_size"] is not None:
            msg += "Note, orig_dest_node_size is replaced by the orig_dest_size parameter, use that instead. "

        warnings.warn(msg)


def plot_shape(
    gdf,
    fc="#cbe0f0",
    ec="#999999",
    linewidth=1,
    alpha=1,
    figsize=(6, 6),
    margin=0.02,
    axis_off=True,
):
    """
    Plot a GeoDataFrame of place boundary geometries.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        the gdf containing the geometries to plot
    fc : string or list
        the facecolor (or list of facecolors) for the polygons
    ec : string or list
        the edgecolor (or list of edgecolors) for the polygons
    linewidth : numeric
        the width of the polygon edge lines
    alpha : numeric
        the opacity
    figsize : tuple
        the size of the plotting figure
    margin : numeric
        the size of the figure margins
    axis_off : bool
        if True, disable the matplotlib axes display

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    warnings.warn(
        "the plot_shape function has been deprecated and will be removed in the next release, use gdf.plot() instead"
    )

    # if facecolor or edgecolor is a string instead of a list, make sure we have
    # as many colors as gdf elements
    if isinstance(fc, str):
        fc = [fc] * len(gdf)
    if isinstance(ec, str):
        ec = [ec] * len(gdf)

    # plot the geometries one at a time
    fig, ax = plt.subplots(figsize=figsize)
    for geometry, facecolor, edgecolor in zip(gdf["geometry"], fc, ec):
        if isinstance(geometry, (Polygon, MultiPolygon)):
            if isinstance(geometry, Polygon):
                geometry = MultiPolygon([geometry])
            for polygon in geometry:
                patch = PolygonPatch(
                    polygon, fc=facecolor, ec=edgecolor, linewidth=linewidth, alpha=alpha
                )
                ax.add_patch(patch)
        else:
            raise ValueError(
                "All geometries in GeoDataFrame must be shapely Polygons or MultiPolygons"
            )

    # adjust the axis margins and limits around the image and make axes
    # equal-aspect
    west, south, east, north = gdf.unary_union.bounds
    margin_ns = (north - south) * margin
    margin_ew = (east - west) * margin
    ax.set_ylim((south - margin_ns, north + margin_ns))
    ax.set_xlim((west - margin_ew, east + margin_ew))
    ax.set_aspect(aspect="equal", adjustable="box")
    if axis_off:
        ax.axis("off")

    plt.show()
    return fig, ax


def _rgb_color_list_to_hex(color_list):
    """
    Convert a list of RGBa colors to a list of hexadecimal color codes.

    Parameters
    ----------
    color_list : list
        the list of RGBa colors

    Returns
    -------
    color_list_hex : list
    """
    color_list_rgb = [[int(x * 255) for x in c[0:3]] for c in color_list]
    color_list_hex = [f"#{rgb[0]:02X}{rgb[1]:02X}{rgb[2]:02X}" for rgb in color_list_rgb]
    return color_list_hex


def get_colors(n, cmap="viridis", start=0.0, stop=1.0, alpha=1.0, return_hex=False):
    """
    Return n-length list of RGBa colors from the passed colormap name and alpha.

    Parameters
    ----------
    n : int
        number of colors
    cmap : string
        name of a colormap
    start : float
        where to start in the colorspace
    stop : float
        where to end in the colorspace
    alpha : float
        opacity, the alpha channel for the RGBa colors
    return_hex : bool
        if True, convert RGBa colors to a hexadecimal string

    Returns
    -------
    colors : list
    """
    colors = [cm.get_cmap(cmap)(x) for x in np.linspace(start, stop, n)]
    colors = [(r, g, b, alpha) for r, g, b, _ in colors]
    if return_hex:
        colors = _rgb_color_list_to_hex(colors)
    return colors


def get_node_colors_by_attr(
    G, attr, num_bins=None, cmap="viridis", start=0, stop=1, na_color="none"
):
    """
    Get a list of node colors by binning continuous attribute into quantiles.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    attr : string
        the name of the attribute
    num_bins : int
        how many quantiles (default None assigns each node to its own bin)
    cmap : string
        name of a colormap
    start : float
        where to start in the colorspace
    stop : float
        where to end in the colorspace
    na_color : string
        what color to assign nodes with null attribute values

    Returns
    -------
    node_colors : list
    """
    if num_bins is None:
        num_bins = len(G)
    bin_labels = range(num_bins)
    attr_values = pd.Series([data[attr] for node, data in G.nodes(data=True)])
    cats = pd.qcut(x=attr_values, q=num_bins, labels=bin_labels)
    colors = get_colors(num_bins, cmap, start, stop)
    node_colors = [colors[int(cat)] if pd.notnull(cat) else na_color for cat in cats]
    return node_colors


def get_edge_colors_by_attr(G, attr, num_bins=5, cmap="viridis", start=0, stop=1, na_color="none"):
    """
    Get a list of edge colors by binning continuous attribute into quantiles.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    attr : string
        the name of the continuous-variable attribute
    num_bins : int
        how many quantiles
    cmap : string
        name of a colormap
    start : float
        where to start in the colorspace
    stop : float
        where to end in the colorspace
    na_color : string
        what color to assign nodes with null attribute values

    Returns
    -------
    edge_colors : list
    """
    if num_bins is None:
        num_bins = len(G.edges())
    bin_labels = range(num_bins)
    attr_values = pd.Series([data[attr] for u, v, key, data in G.edges(keys=True, data=True)])
    cats = pd.qcut(x=attr_values, q=num_bins, labels=bin_labels)
    colors = get_colors(num_bins, cmap, start, stop)
    edge_colors = [colors[int(cat)] if pd.notnull(cat) else na_color for cat in cats]
    return edge_colors


def _save_and_show(fig, ax, save, show, close, filename, file_format, dpi, axis_off):
    """
    Save a figure to disk and show it, as specified.

    Parameters
    ----------
    fig : figure
        matplotlib figure
    ax : axis
        matplotlib axis
    save : bool
        whether to save the figure to disk or not
    show : bool
        whether to display the figure or not
    close : bool
        close the figure (only if show equals False) to prevent display
    filename : string
        the name of the file to save
    file_format : string
        the format of the file to save (e.g., 'jpg', 'png', 'svg')
    dpi : int
        the resolution of the image file if saving
    axis_off : bool
        if True matplotlib axis was turned off by plot_graph so constrain the
        saved figure's extent to the interior of the axis

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    # save the figure if specified
    if save:
        # create the save folder if it doesn't already exist
        if not os.path.exists(settings.imgs_folder):
            os.makedirs(settings.imgs_folder)
        path_filename = os.path.join(settings.imgs_folder, os.extsep.join([filename, file_format]))

        if file_format == "svg":
            # if the file_format is svg, prep the fig/ax a bit for saving
            ax.axis("off")
            ax.set_position([0, 0, 1, 1])
            ax.patch.set_alpha(0.0)
            fig.patch.set_alpha(0.0)
            fig.savefig(
                path_filename,
                bbox_inches=0,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        else:
            if axis_off:
                # if axis is turned off, constrain the saved figure's extent to
                # the interior of the axis
                extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            else:
                extent = "tight"
            fig.savefig(
                path_filename,
                dpi=dpi,
                bbox_inches=extent,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        utils.log("Saved the figure to disk")

    # show the figure if specified
    if show:
        plt.show()
        utils.log("Showed the plot")
    # if show=False, close the figure if close=True to prevent display
    elif close:
        plt.close()

    return fig, ax


def plot_graph(
    G,
    bbox=None,
    fig_height=None,
    fig_width=None,
    margin=0.02,
    axis_off=None,
    equal_aspect=None,
    bgcolor="w",
    show=True,
    save=False,
    close=True,
    file_format=None,
    filename=None,
    dpi=300,
    annotate=None,
    node_color="#66ccff",
    node_size=15,
    node_alpha=1,
    node_edgecolor="none",
    node_zorder=1,
    edge_color="#999999",
    edge_linewidth=1,
    edge_alpha=1,
    use_geom=None,
    figsize=None,
    filepath=None,
):
    """
    Plot a networkx spatial graph.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    bbox : tuple
        bounding box as north,south,east,west - if None will calculate from
        spatial extents of data. if passing a bbox, you probably also want to
        pass margin=0 to constrain it.
    fig_height : int
        deprecated, do not use
    fig_width : int
        deprecated, do not use
    margin : float
        relative margin around the figure
    axis_off : bool
        deprecated, do not use
    equal_aspect : bool
        deprecated, do not use
    bgcolor : string
        the background color of the figure and axis
    show : bool
        if True, show the figure
    save : bool
        if True, save the figure as an image file to disk
    close : bool
        close the figure (only if show equals False) to prevent display
    file_format : string
        deprecated, do not use
    filename : string
        deprecated, do not use
    dpi : int
        the resolution of the image file if saving
    annotate : bool
        deprecated, do not use
    node_color : string
        the color of the nodes. color is passed to matplotlib
    node_size : int
        the size of the nodes
    node_alpha : float
        the opacity of the nodes. if you passed RGBA values to node_color,
        then set this to None to use the alpha channel in node_color
    node_edgecolor : string
        the color of the node's marker's border
    node_zorder : int
        zorder to plot nodes, edges are always 2, so make node_zorder 1 to plot
        nodes beneath them or 3 to plot nodes atop them
    edge_color : string
        the color of the edges' lines. color is passed to matplotlib.
    edge_linewidth : float
        the width of the edges' lines
    edge_alpha : float
        the opacity of the edges' lines. if you passed RGBA values to edge_color,
        then set this to None to use the alpha channel in edge_color
    use_geom : bool
        deprecated, do not use
    figsize : tuple
        figure (width, height)
    filepath : string
        filename.ext to save image in settings.imgs_folder

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    _warn_deprecated_params(
        fig_height=fig_height,
        fig_width=fig_width,
        file_format=file_format,
        filename=filename,
        annotate=annotate,
        use_geom=use_geom,
        axis_off=axis_off,
        equal_aspect=equal_aspect,
    )
    if axis_off is None:
        axis_off = True
    if equal_aspect is None:
        equal_aspect = False
    if fig_height is None:
        fig_height = 6
    if file_format is None:
        file_format = "png"
    if filename is None:
        filename = "temp"
    if annotate is None:
        annotate = False
    if use_geom is None:
        use_geom = True
    if figsize is not None:
        fig_width, fig_height = figsize
    if filepath is not None:
        folders, filename_ext = os.path.split(filepath)
        filename, file_format = os.path.splitext(filename_ext)
        file_format = file_format.strip(".")

    utils.log("Begin plotting the graph...")
    node_Xs = [float(x) for _, x in G.nodes(data="x")]
    node_Ys = [float(y) for _, y in G.nodes(data="y")]

    # get north, south, east, west values either from bbox parameter or from the
    # spatial extent of the edges' geometries
    if bbox is None:
        edges = utils_graph.graph_to_gdfs(G, nodes=False, fill_edge_geometry=True)
        west, south, east, north = edges.total_bounds
    else:
        north, south, east, west = bbox

    # if caller did not pass in a fig_width, calculate it proportionately from
    # the fig_height and bounding box aspect ratio
    bbox_aspect_ratio = (north - south) / (east - west)
    if fig_width is None:
        fig_width = fig_height / bbox_aspect_ratio

    # create the figure and axis
    fig, ax = plt.subplots(figsize=(fig_width, fig_height), facecolor=bgcolor)
    ax.set_facecolor(bgcolor)

    # draw the edges as lines from node to node
    lines = []
    for u, v, data in G.edges(keys=False, data=True):
        if "geometry" in data and use_geom:
            # if it has a geometry attribute (a list of line segments), add them
            # to the list of lines to plot
            xs, ys = data["geometry"].xy
            lines.append(list(zip(xs, ys)))
        else:
            # if it doesn't have a geometry attribute, the edge is a straight
            # line from node to node
            x1 = G.nodes[u]["x"]
            y1 = G.nodes[u]["y"]
            x2 = G.nodes[v]["x"]
            y2 = G.nodes[v]["y"]
            line = [(x1, y1), (x2, y2)]
            lines.append(line)

    # add the lines to the axis as a linecollection
    lc = LineCollection(
        lines, colors=edge_color, linewidths=edge_linewidth, alpha=edge_alpha, zorder=2
    )
    ax.add_collection(lc)
    utils.log("Drew the graph edges")

    # scatter plot the nodes
    ax.scatter(
        node_Xs,
        node_Ys,
        s=node_size,
        c=node_color,
        alpha=node_alpha,
        edgecolor=node_edgecolor,
        zorder=node_zorder,
    )

    # set the extent of the figure
    margin_ns = (north - south) * margin
    margin_ew = (east - west) * margin
    ax.set_ylim((south - margin_ns, north + margin_ns))
    ax.set_xlim((west - margin_ew, east + margin_ew))

    # configure axis appearance
    xaxis = ax.get_xaxis()
    yaxis = ax.get_yaxis()

    xaxis.get_major_formatter().set_useOffset(False)
    yaxis.get_major_formatter().set_useOffset(False)

    # if axis_off, turn off the axis display set the margins to zero and point
    # the ticks in so there's no space around the plot
    if axis_off:
        ax.axis("off")
        ax.margins(0)
        ax.tick_params(which="both", direction="in")
        xaxis.set_visible(False)
        yaxis.set_visible(False)
        fig.canvas.draw()

    if equal_aspect:
        # make everything square
        ax.set_aspect("equal")
        fig.canvas.draw()
    else:
        # if the graph is not projected, conform the aspect ratio to not stretch the plot
        if G.graph["crs"] == settings.default_crs:
            coslat = np.cos((min(node_Ys) + max(node_Ys)) / 2.0 / 180.0 * np.pi)
            ax.set_aspect(1.0 / coslat)
            fig.canvas.draw()

    # annotate the axis with node IDs if annotate=True
    if annotate:
        for node, data in G.nodes(data=True):
            ax.annotate(node, xy=(data["x"], data["y"]))

    # save and show the figure as specified
    fig, ax = _save_and_show(fig, ax, save, show, close, filename, file_format, dpi, axis_off)
    return fig, ax


def _node_list_to_coordinate_lines(G, route, use_geom=True):
    """
    Make list of lines that follow the route defined by the list of nodes.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    route : list
        a valid graph path as a list of node IDs
    use_geom : bool
        if True, use the spatial geometry attribute of the edges to draw
        geographically accurate edges, rather than just lines straight from
        node to node

    Returns
    -------
    lines : list
        list of lines as pairs ((x_start, y_start), (x_stop, y_stop))
    """
    edge_nodes = list(zip(route[:-1], route[1:]))
    lines = []
    for u, v in edge_nodes:
        # if there are parallel edges, select the shortest in length
        data = min(G.get_edge_data(u, v).values(), key=lambda x: x["length"])

        # if it has a geometry attribute (ie, a list of line segments)
        if "geometry" in data and use_geom:
            # add them to the list of lines to plot
            xs, ys = data["geometry"].xy
            lines.append(list(zip(xs, ys)))
        else:
            # if it doesn't have a geometry attribute, the edge is a straight
            # line from node to node
            x1 = G.nodes[u]["x"]
            y1 = G.nodes[u]["y"]
            x2 = G.nodes[v]["x"]
            y2 = G.nodes[v]["y"]
            line = [(x1, y1), (x2, y2)]
            lines.append(line)
    return lines


def plot_graph_route(
    G,
    route,
    bbox=None,
    fig_height=None,
    fig_width=None,
    margin=0.02,
    bgcolor="w",
    axis_off=None,
    show=True,
    save=False,
    close=True,
    file_format=None,
    filename=None,
    dpi=300,
    annotate=None,
    node_color="#999999",
    node_size=15,
    node_alpha=1,
    node_edgecolor="none",
    node_zorder=1,
    edge_color="#999999",
    edge_linewidth=1,
    edge_alpha=1,
    use_geom=None,
    origin_point=None,
    destination_point=None,
    route_color="r",
    route_linewidth=4,
    route_alpha=0.5,
    orig_dest_node_alpha=None,
    orig_dest_node_size=None,
    orig_dest_node_color=None,
    orig_dest_point_color=None,
    figsize=None,
    filepath=None,
    orig_dest_size=None,
):
    """
    Plot a route along a networkx spatial graph.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    route : list
        the route as a list of nodes
    bbox : tuple
        bounding box as north,south,east,west - if None will calculate from
        spatial extents of data. if passing a bbox, you probably also want to
        pass margin=0 to constrain it.
    fig_height : int
        deprecated, do not use
    fig_width : int
        deprecated, do not use
    margin : float
        relative margin around the figure
    axis_off : bool
        deprecated, do not use
    bgcolor : string
        the background color of the figure and axis
    show : bool
        if True, show the figure
    save : bool
        if True, save the figure as an image file to disk
    close : bool
        close the figure (only if show equals False) to prevent display
    file_format : string
        deprecated, do not use
    filename : string
        deprecated, do not use
    dpi : int
        the resolution of the image file if saving
    annotate : bool
        deprecated, do not use
    node_color : string
        the color of the nodes
    node_size : int
        the size of the nodes
    node_alpha : float
        the opacity of the nodes
    node_edgecolor : string
        the color of the node's marker's border
    node_zorder : int
        zorder to plot nodes, edges are always 2, so make node_zorder 1 to plot
        nodes beneath them or 3 to plot nodes atop them
    edge_color : string
        the color of the edges' lines
    edge_linewidth : float
        the width of the edges' lines
    edge_alpha : float
        the opacity of the edges' lines
    use_geom : bool
        deprecated, do not use
    origin_point : tuple
        deprecated, do not use
    destination_point : tuple
        deprecated, do not use
    route_color : string
        the color of the route
    route_linewidth : int
        the width of the route line
    route_alpha : float
        the opacity of the route line
    orig_dest_node_alpha : float
        deprecated, do not use
    orig_dest_node_size : int
        deprecated, do not use
    orig_dest_node_color : string
        deprecated, do not use
    orig_dest_point_color : string
        deprecated, do not use
    figsize : tuple
        figure (width, height)
    filepath : string
        filename.ext to save image in settings.imgs_folder
    orig_dest_size : int
        the size of the origin and destination nodes

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    _warn_deprecated_params(
        fig_height=fig_height,
        fig_width=fig_width,
        file_format=file_format,
        filename=filename,
        annotate=annotate,
        use_geom=use_geom,
        origin_point=origin_point,
        destination_point=destination_point,
        orig_dest_node_alpha=orig_dest_node_alpha,
        orig_dest_node_color=orig_dest_node_color,
        orig_dest_point_color=orig_dest_point_color,
        orig_dest_node_size=orig_dest_node_size,
        axis_off=axis_off,
    )
    if axis_off is None:
        axis_off = True
    if fig_height is None:
        fig_height = 6
    if file_format is None:
        file_format = "png"
    if filename is None:
        filename = "temp"
    if annotate is None:
        annotate = False
    if use_geom is None:
        use_geom = True
    if orig_dest_node_alpha is None:
        orig_dest_node_alpha = 0.5
    if orig_dest_node_size is None:
        orig_dest_node_size = 100
    if orig_dest_node_color is None:
        orig_dest_node_color = "r"
    if orig_dest_point_color is None:
        orig_dest_point_color = "b"
    if figsize is not None:
        fig_width, fig_height = figsize
    if filepath is not None:
        folders, filename_ext = os.path.split(filepath)
        filename, file_format = os.path.splitext(filename_ext)
        file_format = file_format.strip(".")
    if orig_dest_size is not None:
        orig_dest_node_size = orig_dest_size

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # plot the graph but not the route
        fig, ax = plot_graph(
            G,
            bbox=bbox,
            fig_height=fig_height,
            fig_width=fig_width,
            margin=margin,
            axis_off=axis_off,
            bgcolor=bgcolor,
            show=False,
            save=False,
            close=False,
            filename=filename,
            dpi=dpi,
            annotate=annotate,
            node_color=node_color,
            node_size=node_size,
            node_alpha=node_alpha,
            node_edgecolor=node_edgecolor,
            node_zorder=node_zorder,
            edge_color=edge_color,
            edge_linewidth=edge_linewidth,
            edge_alpha=edge_alpha,
            use_geom=use_geom,
        )

    # the origin and destination nodes are the first and last nodes in the route
    origin_node = route[0]
    destination_node = route[-1]

    if origin_point is None or destination_point is None:
        # if caller didn't pass points, use the first and last node in route as
        # origin/destination
        origin_destination_lats = (G.nodes[origin_node]["y"], G.nodes[destination_node]["y"])
        origin_destination_lons = (G.nodes[origin_node]["x"], G.nodes[destination_node]["x"])
    else:
        # otherwise, use the passed points as origin/destination
        origin_destination_lats = (origin_point[0], destination_point[0])
        origin_destination_lons = (origin_point[1], destination_point[1])
        orig_dest_node_color = orig_dest_point_color

    # scatter the origin and destination points
    ax.scatter(
        origin_destination_lons,
        origin_destination_lats,
        s=orig_dest_node_size,
        c=orig_dest_node_color,
        alpha=orig_dest_node_alpha,
        edgecolor=node_edgecolor,
        zorder=4,
    )

    # plot the route lines
    lines = _node_list_to_coordinate_lines(G, route, use_geom)

    # add the lines to the axis as a linecollection
    lc = LineCollection(
        lines, colors=route_color, linewidths=route_linewidth, alpha=route_alpha, zorder=3
    )
    ax.add_collection(lc)

    # save and show the figure as specified
    fig, ax = _save_and_show(fig, ax, save, show, close, filename, file_format, dpi, axis_off)
    return fig, ax


def plot_graph_routes(
    G,
    routes,
    bbox=None,
    fig_height=None,
    fig_width=None,
    margin=0.02,
    bgcolor="w",
    axis_off=None,
    show=True,
    save=False,
    close=True,
    file_format=None,
    filename=None,
    dpi=300,
    annotate=None,
    node_color="#999999",
    node_size=15,
    node_alpha=1,
    node_edgecolor="none",
    node_zorder=1,
    edge_color="#999999",
    edge_linewidth=1,
    edge_alpha=1,
    use_geom=None,
    orig_dest_points=None,
    route_color="r",
    route_linewidth=4,
    route_alpha=0.5,
    orig_dest_node_alpha=None,
    orig_dest_node_size=None,
    orig_dest_node_color=None,
    orig_dest_point_color=None,
    figsize=None,
    filepath=None,
    orig_dest_size=None,
):
    """
    Plot several routes along a networkx spatial graph.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    routes : list
        the routes as a list of lists of nodes
    bbox : tuple
        bounding box as north,south,east,west - if None will calculate from
        spatial extents of data. if passing a bbox, you probably also want to
        pass margin=0 to constrain it.
    fig_height : int
        deprecated, do not use
    fig_width : int
        deprecated, do not use
    margin : float
        relative margin around the figure
    axis_off : bool
        if True turn off the matplotlib axis
    bgcolor : string
        the background color of the figure and axis
    show : bool
        if True, show the figure
    save : bool
        if True, save the figure as an image file to disk
    close : bool
        close the figure (only if show equals False) to prevent display
    file_format : string
        deprecated, do not use
    filename : string
        deprecated, do not use
    dpi : int
        the resolution of the image file if saving
    annotate : bool
        deprecated, do not use
    node_color : string
        the color of the nodes
    node_size : int
        the size of the nodes
    node_alpha : float
        the opacity of the nodes
    node_edgecolor : string
        the color of the node's marker's border
    node_zorder : int
        zorder to plot nodes, edges are always 2, so make node_zorder 1 to plot
        nodes beneath them or 3 to plot nodes atop them
    edge_color : string
        the color of the edges' lines
    edge_linewidth : float
        the width of the edges' lines
    edge_alpha : float
        the opacity of the edges' lines
    use_geom : bool
        deprecated, do not use
    orig_dest_points : list of tuples
        deprecated, do not use
    route_color : string
        route color (note: will be renamed `route_colors` and take list in
        next release)
    route_linewidth : int
        the width of the route line
    route_alpha : float
        the opacity of the route line
    orig_dest_node_alpha : float
        deprecated, do not use
    orig_dest_node_size : int
        deprecated, do not use
    orig_dest_node_color : string
        deprecated, do not use
    orig_dest_point_color : string
        deprecated, do not use
    figsize : tuple
        figure (width, height)
    filepath : string
        filename.ext to save image in settings.imgs_folder
    orig_dest_size : int
        the size of the origin and destination nodes

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    _warn_deprecated_params(
        fig_height=fig_height,
        fig_width=fig_width,
        file_format=file_format,
        filename=filename,
        annotate=annotate,
        use_geom=use_geom,
        orig_dest_points=orig_dest_points,
        orig_dest_node_alpha=orig_dest_node_alpha,
        orig_dest_node_color=orig_dest_node_color,
        orig_dest_point_color=orig_dest_point_color,
        orig_dest_node_size=orig_dest_node_size,
        axis_off=axis_off,
    )
    if axis_off is None:
        axis_off = True
    if fig_height is None:
        fig_height = 6
    if file_format is None:
        file_format = "png"
    if filename is None:
        filename = "temp"
    if annotate is None:
        annotate = False
    if use_geom is None:
        use_geom = True
    if orig_dest_node_alpha is None:
        orig_dest_node_alpha = 0.5
    if orig_dest_node_size is None:
        orig_dest_node_size = 100
    if orig_dest_node_color is None:
        orig_dest_node_color = "r"
    if orig_dest_point_color is None:
        orig_dest_point_color = "b"
    if figsize is not None:
        fig_width, fig_height = figsize
    if filepath is not None:
        folders, filename_ext = os.path.split(filepath)
        filename, file_format = os.path.splitext(filename_ext)
        file_format = file_format.strip(".")
    if orig_dest_size is not None:
        orig_dest_node_size = orig_dest_size

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # plot the graph but not the routes
        fig, ax = plot_graph(
            G,
            bbox=bbox,
            fig_height=fig_height,
            fig_width=fig_width,
            margin=margin,
            axis_off=axis_off,
            bgcolor=bgcolor,
            show=False,
            save=False,
            close=False,
            filename=filename,
            dpi=dpi,
            annotate=annotate,
            node_color=node_color,
            node_size=node_size,
            node_alpha=node_alpha,
            node_edgecolor=node_edgecolor,
            node_zorder=node_zorder,
            edge_color=edge_color,
            edge_linewidth=edge_linewidth,
            edge_alpha=edge_alpha,
            use_geom=use_geom,
        )

    # save coordinates of the given reference points
    orig_dest_points_lats = []
    orig_dest_points_lons = []

    if orig_dest_points is None:
        # if caller didn't pass points, use the first and last node in each route as
        # origin/destination points
        for route in routes:
            origin_node = route[0]
            destination_node = route[-1]
            orig_dest_points_lats.append(G.nodes[origin_node]["y"])
            orig_dest_points_lats.append(G.nodes[destination_node]["y"])
            orig_dest_points_lons.append(G.nodes[origin_node]["x"])
            orig_dest_points_lons.append(G.nodes[destination_node]["x"])

    else:
        # otherwise, use the passed points as origin/destination points
        for point in orig_dest_points:
            orig_dest_points_lats.append(point[0])
            orig_dest_points_lons.append(point[1])
        orig_dest_node_color = orig_dest_point_color

    # scatter the origin and destination points
    ax.scatter(
        orig_dest_points_lons,
        orig_dest_points_lats,
        s=orig_dest_node_size,
        c=orig_dest_node_color,
        alpha=orig_dest_node_alpha,
        edgecolor=node_edgecolor,
        zorder=4,
    )

    # plot the routes lines
    lines = []
    for route in routes:
        lines.extend(_node_list_to_coordinate_lines(G, route, use_geom))

    # add the lines to the axis as a linecollection
    lc = LineCollection(
        lines, colors=route_color, linewidths=route_linewidth, alpha=route_alpha, zorder=3
    )
    ax.add_collection(lc)

    # save and show the figure as specified
    fig, ax = _save_and_show(fig, ax, save, show, close, filename, file_format, dpi, axis_off)
    return fig, ax


def plot_figure_ground(
    G=None,
    address=None,
    point=None,
    dist=805,
    network_type="drive_service",
    street_widths=None,
    default_width=4,
    fig_length=None,
    edge_color="w",
    bgcolor="#333333",
    smooth_joints=True,
    filename=None,
    file_format=None,
    show=False,
    save=True,
    close=True,
    dpi=300,
    figsize=None,
    filepath=None,
):
    """
    Plot figure-ground diagram of a street network.

    Defaults to one square mile.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph, must be unprojected
    address : string
        the address to geocode as the center point if G is not passed in
    point : tuple
        the center point if address and G are not passed in
    dist : numeric
        how many meters to extend north, south, east, and west from the center
        point
    network_type : string
        what type of network to get
    street_widths : dict
        where keys are street types and values are widths to plot in pixels
    default_width : numeric
        the default street width in pixels for any street type not found in
        street_widths dict
    fig_length : numeric
        deprecated, do not use
    edge_color : string
        the color of the streets
    bgcolor : string
        the color of the background
    smooth_joints : bool
        if True, plot nodes same width as streets to smooth line joints and
        prevent cracks between them from showing
    filename : string
        deprecated, do not use
    file_format : string
        deprecated, do not use
    show : bool
        if True, show the figure
    save : bool
        if True, save the figure as an image file to disk
    close : bool
        close the figure (only if show equals False) to prevent display
    dpi : int
        the resolution of the image file if saving
    figsize : tuple
        figure width, height (should be equal to each other)
    filepath : string
        filename.ext to save image in settings.imgs_folder

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    _warn_deprecated_params(fig_length=fig_length, file_format=file_format, filename=filename)
    if fig_length is None:
        fig_length = 8
    if file_format is None:
        file_format = "png"
    if filename is None:
        filename = "temp"
    if figsize is not None:
        fig_length = figsize[0]
    if filepath is not None:
        folders, filename_ext = os.path.split(filepath)
        filename, file_format = os.path.splitext(filename_ext)
        file_format = file_format.strip(".")

    multiplier = 1.2

    # if G was passed-in, use this graph in the plot, centered on the centroid
    # of its nodes
    if G is not None:
        gdf_nodes = utils_graph.graph_to_gdfs(G, edges=False, node_geometry=True)
        lnglat_point = gdf_nodes.unary_union.centroid.coords[0]
        point = tuple(reversed(lnglat_point))

    # otherwise, get the network by either address or point, whichever was
    # passed-in, using a distance multiplier to make sure we get more than
    # enough network. simplify in non-strict mode to not combine multiple street
    # types into single edge
    elif address is not None:
        G, point = graph.graph_from_address(
            address,
            dist=dist * multiplier,
            dist_type="bbox",
            network_type=network_type,
            simplify=False,
            truncate_by_edge=True,
            return_coords=True,
        )
        G = simplification.simplify_graph(G, strict=False)
    elif point is not None:
        G = graph.graph_from_point(
            point,
            dist=dist * multiplier,
            dist_type="bbox",
            network_type=network_type,
            simplify=False,
            truncate_by_edge=True,
        )
        G = simplification.simplify_graph(G, strict=False)
    else:
        raise ValueError("You must pass an address or lat-lng point or graph.")

    # if user did not pass in custom street widths, create a dict of default
    # values
    if street_widths is None:
        street_widths = {
            "footway": 1.5,
            "steps": 1.5,
            "pedestrian": 1.5,
            "service": 1.5,
            "path": 1.5,
            "track": 1.5,
            "motorway": 6,
        }

    # we need an undirected graph to find every edge incident to a node
    G_undir = G.to_undirected()

    # for each network edge, get a linewidth according to street type (the OSM
    # 'highway' value)
    edge_linewidths = []
    for _, _, data in G_undir.edges(keys=False, data=True):
        street_type = data["highway"][0] if isinstance(data["highway"], list) else data["highway"]
        if street_type in street_widths:
            edge_linewidths.append(street_widths[street_type])
        else:
            edge_linewidths.append(default_width)

    if smooth_joints:
        # for each node, get a nodesize according to the narrowest incident edge
        node_widths = {}
        for node in G_undir.nodes():
            # first, identify all the highway types of this node's incident edges
            incident_edges_data = [
                G_undir.get_edge_data(node, neighbor) for neighbor in G_undir.neighbors(node)
            ]
            edge_types = [data[0]["highway"] for data in incident_edges_data]
            if len(edge_types) < 1:
                # if node has no incident edges, make size zero
                node_widths[node] = 0
            else:
                # flatten the list of edge types
                edge_types_flat = []
                for et in edge_types:
                    if isinstance(et, list):
                        edge_types_flat.extend(et)
                    else:
                        edge_types_flat.append(et)

                # for each edge type in the flattened list, lookup the
                # corresponding width
                edge_widths = [
                    street_widths[edge_type] if edge_type in street_widths else default_width
                    for edge_type in edge_types_flat
                ]

                # the node diameter will be the biggest of the edge widths, to
                # make joints perfectly smooth alternatively, use min (?) to
                # prevent anything larger from extending past smallest
                # street's line
                circle_diameter = max(edge_widths)

                # mpl circle marker sizes are in area, so it is the diameter
                # squared
                circle_area = circle_diameter ** 2
                node_widths[node] = circle_area

        # assign the node size to each node in the graph
        node_sizes = [node_widths[node] for node in G_undir.nodes()]
    else:
        node_sizes = 0

    # define the spatial extents of the plotting figure to make it square, in
    # projected units, and cropped to the desired area
    bbox = utils_geo.bbox_from_point(point, dist, project_utm=False)

    # create a filename if one was not passed
    if filename is None and save:
        filename = "figure_ground"

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # plot the figure
        fig, ax = plot_graph(
            G_undir,
            bbox=bbox,
            fig_height=fig_length,
            margin=0,
            axis_off=True,
            equal_aspect=False,
            bgcolor=bgcolor,
            node_size=node_sizes,
            node_color=edge_color,
            edge_linewidth=edge_linewidths,
            edge_color=edge_color,
            show=show,
            save=save,
            close=close,
            filename=filename,
            file_format=file_format,
            dpi=dpi,
        )

    return fig, ax


def plot_footprints(
    gdf,
    fig=None,
    ax=None,
    figsize=None,
    color="#333333",
    bgcolor="w",
    set_bounds=True,
    bbox=None,
    save=False,
    show=True,
    close=False,
    filename=None,
    file_format=None,
    dpi=600,
    filepath=None,
):
    """
    Plot a GeoDataFrame of footprints.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame of footprints
    fig : figure
        matplotlib figure
    ax : axis
        matplotlib axis
    figsize : tuple
        (width, height) size of matplotlib figure
    color : string
        the color of the footprints
    bgcolor : string
        the background color of the plot
    set_bounds : bool
        if True, set bounds from either passed-in bbox or the spatial extent of the gdf
    bbox : tuple
        if True and if set_bounds is True, set the display bounds to this bbox
    save : bool
        whether to save the figure to disk or not
    show : bool
        whether to display the figure or not
    close : bool
        close the figure (only if show equals False) to prevent display
    filename : string
        deprecated, do not use
    file_format : string
        deprecated, do not use
    dpi : int
        the resolution of the image file if saving
    filepath : string
        filename.ext to save image in settings.imgs_folder

    Returns
    -------
    fig, ax : tuple
        matplotlib figure, axis
    """
    _warn_deprecated_params(file_format=file_format, filename=filename)
    if file_format is None:
        file_format = "png"
    if filename is None:
        filename = "temp"
    if filepath is not None:
        folders, filename_ext = os.path.split(filepath)
        filename, file_format = os.path.splitext(filename_ext)
        file_format = file_format.strip(".")

    if fig is None or ax is None:
        fig, ax = plt.subplots(figsize=figsize, facecolor=bgcolor)
        ax.set_facecolor(bgcolor)

    # extract each polygon as a descartes patch, and add to a matplotlib patch
    # collection
    patches = []
    for geometry in gdf["geometry"]:
        if isinstance(geometry, Polygon):
            patches.append(PolygonPatch(geometry))
        elif isinstance(geometry, MultiPolygon):
            for (
                subpolygon
            ) in geometry:  # if geometry is multipolygon, go through each constituent subpolygon
                patches.append(PolygonPatch(subpolygon))
    pc = PatchCollection(patches, facecolor=color, edgecolor=color, linewidth=0, alpha=1)
    ax.add_collection(pc)

    if set_bounds:
        if bbox is None:
            # set the figure bounds to the polygons' bounds
            left, bottom, right, top = gdf.total_bounds
        else:
            top, bottom, right, left = bbox
        ax.set_xlim((left, right))
        ax.set_ylim((bottom, top))

    # turn off the axis display set the margins to zero and point the ticks in
    # so there's no space around the plot
    ax.axis("off")
    ax.margins(0)
    ax.tick_params(which="both", direction="in")
    fig.canvas.draw()

    # make everything square
    ax.set_aspect("equal")
    fig.canvas.draw()

    fig, ax = _save_and_show(
        fig=fig,
        ax=ax,
        save=save,
        show=show,
        close=close,
        filename=filename,
        file_format=file_format,
        dpi=dpi,
        axis_off=True,
    )

    return fig, ax



"""Download points of interests (POIs) from OpenStreetMap."""

import geopandas as gpd
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon

from . import boundaries
from . import downloader
from . import settings
from . import utils
from . import utils_geo


def _create_poi_query(polygon, tags):
    """
    Create an overpass query string based on passed tags.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
        geographic boundaries to fetch POIs within
    tags : dict
        dict of tags used for finding POIs from the selected area

    Returns
    -------
    query : string
    """
    overpass_settings = downloader._make_overpass_settings()

    # make sure every value in dict is bool, str, or list of str
    error_msg = "tags must be a dict with values of bool, str, or list of str"
    if not isinstance(tags, dict):
        raise TypeError(error_msg)

    tags_dict = {}
    for key, value in tags.items():

        if isinstance(value, bool):
            tags_dict[key] = value

        elif isinstance(value, str):
            tags_dict[key] = [value]

        elif isinstance(value, list):
            if not all(isinstance(s, str) for s in value):
                raise TypeError(error_msg)
            tags_dict[key] = value

        else:
            raise TypeError(error_msg)

    # convert the tags dict into a list of {tag:value} dicts
    tags_list = []
    for key, value in tags_dict.items():
        if isinstance(value, bool):
            tags_list.append({key: value})
        else:
            for value_item in value:
                tags_list.append({key: value_item})

    # create query bounding box
    west, south, east, north = polygon.bounds
    bbox = f"({south:.6f},{west:.6f},{north:.6f},{east:.6f})"

    # add node/way/relation query components one at a time
    components = []
    for d in tags_list:
        for key, value in d.items():

            if isinstance(value, bool):
                # if bool (ie, True) just pass the key, no value
                tag_str = f'["{key}"]{bbox};(._;>;);'
            else:
                # otherwise, pass "key"="value"
                tag_str = f'["{key}"="{value}"]{bbox};(._;>;);'

            for kind in ["node", "way", "relation"]:
                components.append(f"({kind}{tag_str});")

    # finalize query and return
    components = "".join(components)
    query = f"{overpass_settings};({components});out;"

    return query


def _osm_poi_download(polygon, tags):
    """
    Get points of interests (POIs) from OpenStreetMap based on passed tags.

    Note that if a polygon is passed-in, the query will be limited to its
    bounding box rather than to the shape of the polygon itself.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
        geographic boundaries to fetch POIs within
    tags : dict
        dict of tags used for finding POIs from the selected area

    Returns
    -------
    responses : dict
        JSON response with POIs from Overpass server
    """
    # TODO: add functionality for subdividing search area geometry
    # TODO: add functionality for constraining query to poly rather than its bbox

    # get the POIs
    query = _create_poi_query(polygon, tags)
    responses = downloader.overpass_request(data={"data": query})
    return responses


def _parse_nodes_coords(osm_response):
    """
    Parse node coordinates from OSM response.

    Some nodes are standalone points of interest, others are vertices in
    polygonal (areal) POIs.

    Parameters
    ----------
    osm_response : string
        OSM response JSON string

    Returns
    -------
    coords : dict
        dict of node IDs and their lat, lng coordinates
    """
    coords = {}
    for result in osm_response["elements"]:
        if "type" in result and result["type"] == "node":
            coords[result["id"]] = {"lat": result["lat"], "lon": result["lon"]}
    return coords


def _parse_polygonal_poi(coords, response):
    """
    Parse areal POI way polygons from OSM node coords.

    Parameters
    ----------
    coords : dict
        dict of node IDs and their lat, lng coordinates
    response : string
        OSM response JSON string

    Returns
    -------
    poi : dict
        dict of POIs containing each's nodes, polygon geometry, and osmid, or
        None if it cannot
    """
    if "type" in response and response["type"] == "way":
        nodes = response["nodes"]
        try:
            polygon = Polygon([(coords[node]["lon"], coords[node]["lat"]) for node in nodes])

            poi = {"nodes": nodes, "geometry": polygon, "osmid": response["id"]}

            if "tags" in response:
                for tag in response["tags"]:
                    poi[tag] = response["tags"][tag]
            return poi

        except Exception:
            utils.log(f"Polygon has invalid geometry: {nodes}")


def _parse_osm_node(response):
    """
    Parse points from OSM nodes.

    Parameters
    ----------
    response : JSON
        nodes from OSM response

    Returns
    -------
    poi : dict
        dict of vertex IDs and their lat, lng coordinates
    """
    try:
        point = Point(response["lon"], response["lat"])

        poi = {"osmid": response["id"], "geometry": point}

        if "tags" in response:
            for tag in response["tags"]:
                poi[tag] = response["tags"][tag]

    except Exception:
        utils.log(f'Point has invalid geometry: {response["id"]}')

    return poi


def _invalid_multipoly_handler(gdf, relation, way_ids):  # pragma: no cover
    """
    Handle invalid multipolygon geometries.

    For example, when there exists a feature without geometry (geometry==NaN).

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame with Polygon geometries that should be converted into a
        MultiPolygon object.
    relation : dict
        OSM 'relation' dictionary
    way_ids : list
        A list of 'way' ids that should be converted into a MultiPolygon object.

    Returns
    -------
    shapely.geometry.MultiPolygon
    """
    try:
        gdf_clean = gdf.dropna(subset=["geometry"])
        multipoly = MultiPolygon(list(gdf_clean["geometry"]))
        return multipoly

    except Exception:
        utils.log(f'Invalid geometry at relation "{relation["id"]}", way IDs: {way_ids}')
        return None


def _parse_osm_relations(relations, df_osm_ways):
    """
    Parse OSM relations (MultiPolygons) from ways and nodes.

    See more information about relations from OSM documentation:
    http://wiki.openstreetmap.org/wiki/Relation

    Parameters
    ----------
    relations : list
        OSM 'relation' items (dictionaries) in a list
    df_osm_ways : geopandas..GeoDataFrame
        OSM 'way' features as a GeoDataFrame that contains all the 'way'
        features that will constitute the multipolygon relations

    Returns
    -------
    df_osm_ways : geopandas.GeoDataFrame
        GeoDataFrame with MultiPolygon representations of the relations and
        the attributes associated with them.
    """
    gdf_relations = gpd.GeoDataFrame()

    # Iterate over relations and extract the items
    for relation in relations:
        try:
            if relation["tags"]["type"] == "multipolygon":
                # Parse member 'way' ids
                member_way_ids = [
                    member["ref"] for member in relation["members"] if member["type"] == "way"
                ]
                # Extract the ways
                member_ways = df_osm_ways.reindex(member_way_ids)
                # Extract the nodes of those ways
                member_nodes = list(member_ways["nodes"].values)
                try:
                    # Create MultiPolygon from geometries (exclude NaNs)
                    multipoly = MultiPolygon(list(member_ways["geometry"]))
                except Exception:
                    multipoly = _invalid_multipoly_handler(
                        gdf=member_ways, relation=relation, way_ids=member_way_ids
                    )

                if multipoly:
                    # Create GeoDataFrame with the tags and the MultiPolygon and its
                    # 'ways' (ids), and the 'nodes' of those ways
                    geo = gpd.GeoDataFrame(relation["tags"], index=[relation["id"]])
                    # Initialize columns (needed for .loc inserts)
                    geo = geo.assign(
                        geometry=None, ways=None, nodes=None, element_type=None, osmid=None
                    )
                    # Add attributes
                    geo.loc[relation["id"], "geometry"] = multipoly
                    geo.loc[relation["id"], "ways"] = member_way_ids
                    geo.loc[relation["id"], "nodes"] = member_nodes
                    geo.loc[relation["id"], "element_type"] = "relation"
                    geo.loc[relation["id"], "osmid"] = relation["id"]

                    # Append to relation GeoDataFrame
                    gdf_relations = gdf_relations.append(geo, sort=False)
                    # Remove such 'ways' from 'df_osm_ways' that are part of the 'relation'
                    df_osm_ways = df_osm_ways.drop(member_way_ids)
        except Exception:
            utils.log(f'Could not parse OSM relation {relation["id"]}')

    # Merge df_osm_ways and the gdf_relations
    df_osm_ways = df_osm_ways.append(gdf_relations, sort=False)
    return df_osm_ways


def _create_poi_gdf(polygon, tags):
    """
    Create GeoDataFrame from POIs json returned by Overpass API.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
        geographic boundaries to fetch POIs within
    tags : dict
        dict of tags used for finding POIs from the selected area

    Returns
    -------
    gdf : geopandas.GeoDataFrame
        POIs and their associated tags
    """
    responses = _osm_poi_download(polygon, tags)

    # Parse coordinates from all the nodes in the response
    coords = _parse_nodes_coords(responses)

    # POI nodes
    poi_nodes = {}

    # POI ways
    poi_ways = {}

    # A list of POI relations
    relations = []

    for result in responses["elements"]:
        if result["type"] == "node" and "tags" in result:
            poi = _parse_osm_node(response=result)
            # Add element_type
            poi["element_type"] = "node"
            # Add to 'pois'
            poi_nodes[result["id"]] = poi
        elif result["type"] == "way":
            # Parse POI area Polygon
            poi_area = _parse_polygonal_poi(coords=coords, response=result)
            if poi_area:
                # Add element_type
                poi_area["element_type"] = "way"
                # Add to 'poi_ways'
                poi_ways[result["id"]] = poi_area

        elif result["type"] == "relation":
            # Add relation to a relation list (needs to be parsed after
            # all nodes and ways have been parsed)
            relations.append(result)

    # Create GeoDataFrames
    gdf_nodes = gpd.GeoDataFrame(poi_nodes).T
    gdf_nodes.crs = settings.default_crs

    gdf_ways = gpd.GeoDataFrame(poi_ways).T
    gdf_ways.crs = settings.default_crs

    # Parse relations (MultiPolygons) from 'ways'
    gdf_ways = _parse_osm_relations(relations=relations, df_osm_ways=gdf_ways)

    # Combine GeoDataFrames
    gdf = gdf_nodes.append(gdf_ways, sort=False)

    # if caller requested pois within a polygon, only retain those that
    # fall within the polygon
    if len(gdf) > 0:
        gdf = gdf.loc[gdf["geometry"].centroid.within(polygon)]

    return gdf


def pois_from_point(point, tags, dist=1000, timeout=None, memory=None, custom_settings=None):
    """
    Get point of interests (POIs) within some distance N, S, E, W of a point.

    Parameters
    ----------
    point : tuple
        a (lat, lng) point
    tags : dict
        Dict of tags used for finding POIs from the selected area. Results
        returned are the union, not intersection of each individual tag.
        Each result matches at least one tag given. The dict keys should be
        OSM tags, (e.g., `amenity`, `landuse`, `highway`, etc) and the dict
        values should be either `True` to retrieve all items with the given
        tag, or a string to get a single tag-value combination, or a list of
        strings to get multiple values for the given tag. For example,
        `tags = {'amenity':True, 'landuse':['retail','commercial'],
        'highway':'bus_stop'}` would return all amenities, landuse=retail,
        landuse=commercial, and highway=bus_stop.
    dist : numeric
        distance in meters
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    north, south, east, west = utils_geo.bbox_from_point(point=point, dist=dist)
    polygon = utils_geo.bbox_to_poly(north, south, east, west)
    return pois_from_polygon(
        polygon, tags, timeout=timeout, memory=memory, custom_settings=custom_settings
    )


def pois_from_address(address, tags, dist=1000, timeout=None, memory=None, custom_settings=None):
    """
    Get point of interests (POIs) within some distance N, S, E, W of address.

    Parameters
    ----------
    address : string
        the address to geocode to a lat-lng point
    tags : dict
        Dict of tags used for finding POIs from the selected area. Results
        returned are the union, not intersection of each individual tag.
        Each result matches at least one tag given. The dict keys should be
        OSM tags, (e.g., `amenity`, `landuse`, `highway`, etc) and the dict
        values should be either `True` to retrieve all items with the given
        tag, or a string to get a single tag-value combination, or a list of
        strings to get multiple values for the given tag. For example,
        `tags = {'amenity':True, 'landuse':['retail','commercial'],
        'highway':'bus_stop'}` would return all amenities, landuse=retail,
        landuse=commercial, and highway=bus_stop.
    dist : numeric
        distance in meters
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    # geocode the address string to a (lat, lng) point
    point = utils_geo.geocode(query=address)
    return pois_from_point(
        point=point,
        tags=tags,
        dist=dist,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
    )


def pois_from_place(place, tags, which_result=1, timeout=None, memory=None, custom_settings=None):
    """
    Get points of interest (POIs) within the boundaries of some place.

    Parameters
    ----------
    place : string
        the query to geocode to get boundary polygon
    tags : dict
        Dict of tags used for finding POIs from the selected area. Results
        returned are the union, not intersection of each individual tag.
        Each result matches at least one tag given. The dict keys should be
        OSM tags, (e.g., `amenity`, `landuse`, `highway`, etc) and the dict
        values should be either `True` to retrieve all items with the given
        tag, or a string to get a single tag-value combination, or a list of
        strings to get multiple values for the given tag. For example,
        `tags = {'amenity':True, 'landuse':['retail','commercial'],
        'highway':'bus_stop'}` would return all amenities, landuse=retail,
        landuse=commercial, and highway=bus_stop.
    which_result : int
        max number of geocoding results to return and which to process
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    city = boundaries.gdf_from_place(place, which_result=which_result)
    polygon = city["geometry"].iloc[0]
    return pois_from_polygon(
        polygon, tags, timeout=timeout, memory=memory, custom_settings=custom_settings
    )


def pois_from_polygon(polygon, tags, timeout=None, memory=None, custom_settings=None):
    """
    Get point of interests (POIs) within some polygon.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
        geographic boundaries to fetch POIs within
    tags : dict
        Dict of tags used for finding POIs from the selected area. Results
        returned are the union, not intersection of each individual tag.
        Each result matches at least one tag given. The dict keys should be
        OSM tags, (e.g., `amenity`, `landuse`, `highway`, etc) and the dict
        values should be either `True` to retrieve all items with the given
        tag, or a string to get a single tag-value combination, or a list of
        strings to get multiple values for the given tag. For example,
        `tags = {'amenity':True, 'landuse':['retail','commercial'],
        'highway':'bus_stop'}` would return all amenities, landuse=retail,
        landuse=commercial, and highway=bus_stop.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    utils._handle_deprecated_params(timeout=timeout, memory=memory, custom_settings=custom_settings)
    return _create_poi_gdf(polygon, tags)



"""Create GeoDataFrames of place boundaries."""

import logging as lg

import geopandas as gpd

from . import downloader
from . import projection
from . import settings
from . import utils


def gdf_from_place(query, which_result=1, buffer_dist=None):
    """
    Create a GeoDataFrame from a single place name query.

    Geocode the query with Nominatim then turn it into a GeoDataFrame with
    a geometry column.

    Parameters
    ----------
    query : string or dict
        query string or structured query dict to geocode/download
    which_result : int
        max number of results to return and which to process upon receipt
    buffer_dist : float
        distance to buffer around the place geometry, in meters

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    # ensure query type
    if not isinstance(query, (str, dict)):
        raise ValueError("query must be a dict or a string")

    # get the data from OSM
    data = downloader._osm_polygon_download(query, limit=which_result)
    if len(data) >= which_result:

        # extract data elements from the JSON response
        result = data[which_result - 1]
        bbox_south, bbox_north, bbox_west, bbox_east = [float(x) for x in result["boundingbox"]]
        geometry = result["geojson"]
        place = result["display_name"]
        features = [
            {
                "type": "Feature",
                "geometry": geometry,
                "properties": {
                    "place_name": place,
                    "bbox_north": bbox_north,
                    "bbox_south": bbox_south,
                    "bbox_east": bbox_east,
                    "bbox_west": bbox_west,
                },
            }
        ]

        # if we got an unexpected geometry type (like a point), log a warning
        if geometry["type"] not in ["Polygon", "MultiPolygon"]:
            utils.log(f'OSM returned a {geometry["type"]} as the geometry', level=lg.WARNING)

        # create the GeoDataFrame, name it, and set its original CRS to default_crs
        gdf = gpd.GeoDataFrame.from_features(features)
        gdf.crs = settings.default_crs

        # if buffer_dist was passed in, project the geometry to UTM, buffer it
        # in meters, then project it back to lat-lng
        if buffer_dist is not None:
            gdf_utm = projection.project_gdf(gdf)
            gdf_utm["geometry"] = gdf_utm["geometry"].buffer(buffer_dist)
            gdf = projection.project_gdf(gdf_utm, to_latlong=True)
            utils.log(f"Buffered GeoDataFrame to {buffer_dist} meters")

        # return the gdf
        utils.log(f'Created GeoDataFrame with {len(gdf)} row for query "{query}"')
        return gdf
    else:
        # if no data returned (or fewer results than which_result)
        utils.log(
            f'OSM returned no results (or fewer than which_result) for query "{query}"',
            level=lg.WARNING,
        )
        return gpd.GeoDataFrame()


def gdf_from_places(queries, which_results=None, buffer_dist=None):
    """
    Create a GeoDataFrame from a list of place name queries.

    Geocode the queries with Nominatim then turn result into GeoDataFrame with
    a geometry column.

    Parameters
    ----------
    queries : list
        list of query strings or structured query dicts to geocode/download,
        one at a time
    which_results : list
        if not None, a list of max number of results to return and which to
        process upon receipt, for each query in queries
    buffer_dist : float
        distance to buffer around the place geometry, in meters

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    # create an empty GeoDataFrame then append each result as a new row,
    # checking for the presence of which_results
    gdf = gpd.GeoDataFrame()
    if which_results is not None:

        if len(queries) != len(which_results):
            raise ValueError("which_results length must equal queries length")

        for query, which_result in zip(queries, which_results):
            gdf_tmp = gdf_from_place(query, buffer_dist=buffer_dist, which_result=which_result)
            gdf = gdf.append(gdf_tmp)
    else:
        for query in queries:
            gdf = gdf.append(gdf_from_place(query, buffer_dist=buffer_dist))

    # reset the index
    gdf = gdf.reset_index(drop=True)

    # set the original CRS of the GeoDataFrame to default_crs, and return it
    gdf.crs = settings.default_crs
    utils.log(f"Finished creating GeoDataFrame with {len(gdf)} rows from {len(queries)} queries")
    return gdf



"""Graph creation functions."""

import bz2
import os
import xml.sax
from itertools import groupby

import networkx as nx
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

from . import boundaries
from . import distance
from . import downloader
from . import projection
from . import settings
from . import simplification
from . import truncate
from . import utils
from . import utils_geo
from . import utils_graph
from ._errors import EmptyOverpassResponse
from ._version import __version__


def graph_from_bbox(
    north,
    south,
    east,
    west,
    network_type="all_private",
    simplify=True,
    retain_all=False,
    truncate_by_edge=False,
    clean_periphery=True,
    custom_filter=None,
    timeout=None,
    memory=None,
    custom_settings=None,
    max_query_area_size=None,
):
    """
    Create a graph from OSM within some bounding box.

    Parameters
    ----------
    north : float
        northern latitude of bounding box
    south : float
        southern latitude of bounding box
    east : float
        eastern longitude of bounding box
    west : float
        western longitude of bounding box
    network_type : string
        what type of street network to get if custom_filter is None. One of
        'walk', 'bike', 'drive', 'drive_service', 'all', or 'all_private'.
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected
    truncate_by_edge : bool
        if True, retain node if it's outside bounding box but at least one of
        node's neighbors are within the bounding box
    clean_periphery : bool
        if True, buffer 500m to get a graph larger than requested, then
        simplify, then truncate it to requested spatial boundaries
    custom_filter : string
        a custom network filter to be used instead of the network_type presets,
        e.g., '["power"~"line"]' or '["highway"~"motorway|trunk"]'. Also pass
        in a network_type that is in settings.bidirectional_network_types if
        you want graph to be fully bidirectional.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module
    max_query_area_size : None
        deprecated, use ox.config(max_query_area_size=value) instead to
        configure this setting via the settings module

    Returns
    -------
    G : networkx.MultiDiGraph
    """
    # convert bounding box to a polygon
    polygon = utils_geo.bbox_to_poly(north, south, east, west)

    # create graph using this polygon geometry
    G = graph_from_polygon(
        polygon,
        network_type=network_type,
        simplify=simplify,
        retain_all=retain_all,
        truncate_by_edge=truncate_by_edge,
        clean_periphery=clean_periphery,
        custom_filter=custom_filter,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
        max_query_area_size=max_query_area_size,
    )

    utils.log(f"graph_from_bbox returned graph with {len(G)} nodes and {len(G.edges())} edges")
    return G


def graph_from_point(
    center_point,
    dist=1000,
    dist_type="bbox",
    network_type="all_private",
    simplify=True,
    retain_all=False,
    truncate_by_edge=False,
    clean_periphery=True,
    custom_filter=None,
    timeout=None,
    memory=None,
    custom_settings=None,
    max_query_area_size=None,
):
    """
    Create a graph from OSM within some distance of some (lat, lng) point.

    Parameters
    ----------
    center_point : tuple
        the (lat, lng) center point around which to construct the graph
    dist : int
        retain only those nodes within this many meters of the center of the
        graph, with distance determined according to dist_type argument
    dist_type : string
        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding
        box of the distance parameter. if 'network', retain only those nodes
        within some network distance from the center-most node.
    network_type : string
        what type of street network to get if custom_filter is None. One of
        'walk', 'bike', 'drive', 'drive_service', 'all', or 'all_private'.
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected
    truncate_by_edge : bool
        if True, retain node if it's outside bounding box but at least one of
        node's neighbors are within bounding box
    clean_periphery : bool,
        if True, buffer 500m to get a graph larger than requested, then
        simplify, then truncate it to requested spatial boundaries
    custom_filter : string
        a custom network filter to be used instead of the network_type presets,
        e.g., '["power"~"line"]' or '["highway"~"motorway|trunk"]'. Also pass
        in a network_type that is in settings.bidirectional_network_types if
        you want graph to be fully bidirectional.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module
    max_query_area_size : None
        deprecated, use ox.config(max_query_area_size=value) instead to
        configure this setting via the settings module

    Returns
    -------
    G : networkx.MultiDiGraph
    """
    if dist_type not in {"bbox", "network"}:
        raise ValueError('dist_type must be "bbox" or "network"')

    # create a bounding box from the center point and the distance in each
    # direction
    north, south, east, west = utils_geo.bbox_from_point(center_point, dist)

    # create a graph from the bounding box
    G = graph_from_bbox(
        north,
        south,
        east,
        west,
        network_type=network_type,
        simplify=simplify,
        retain_all=retain_all,
        truncate_by_edge=truncate_by_edge,
        clean_periphery=clean_periphery,
        custom_filter=custom_filter,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
        max_query_area_size=max_query_area_size,
    )

    # if the network dist_type is network, find the node in the graph
    # nearest to the center point, and truncate the graph by network distance
    # from this node
    if dist_type == "network":
        centermost_node = distance.get_nearest_node(G, center_point)
        G = truncate.truncate_graph_dist(G, centermost_node, max_dist=dist)

    utils.log(f"graph_from_point returned graph with {len(G)} nodes and {len(G.edges())} edges")
    return G


def graph_from_address(
    address,
    dist=1000,
    dist_type="bbox",
    network_type="all_private",
    simplify=True,
    retain_all=False,
    truncate_by_edge=False,
    return_coords=False,
    clean_periphery=True,
    custom_filter=None,
    timeout=None,
    memory=None,
    custom_settings=None,
    max_query_area_size=None,
):
    """
    Create a graph from OSM within some distance of some address.

    Parameters
    ----------
    address : string
        the address to geocode and use as the central point around which to
        construct the graph
    dist : int
        retain only those nodes within this many meters of the center of the
        graph
    dist_type : string
        {'network', 'bbox'} if 'bbox', retain only those nodes within a bounding
        box of the distance parameter.
        if 'network', retain only those nodes within some network distance from
        the center-most node.
    network_type : string
        what type of street network to get if custom_filter is None. One of
        'walk', 'bike', 'drive', 'drive_service', 'all', or 'all_private'.
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected
    truncate_by_edge : bool
        if True, retain node if it's outside bounding box but at least one of
        node's neighbors are within bounding box
    return_coords : bool
        optionally also return the geocoded coordinates of the address
    clean_periphery : bool,
        if True, buffer 500m to get a graph larger than requested, then
        simplify, then truncate it to requested spatial boundaries
    custom_filter : string
        a custom network filter to be used instead of the network_type presets,
        e.g., '["power"~"line"]' or '["highway"~"motorway|trunk"]'. Also pass
        in a network_type that is in settings.bidirectional_network_types if
        you want graph to be fully bidirectional.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module
    max_query_area_size : None
        deprecated, use ox.config(max_query_area_size=value) instead to
        configure this setting via the settings module

    Returns
    -------
    networkx.MultiDiGraph or optionally (networkx.MultiDiGraph, (lat, lng))
    """
    # geocode the address string to a (lat, lng) point
    point = utils_geo.geocode(query=address)

    # then create a graph from this point
    G = graph_from_point(
        point,
        dist,
        dist_type,
        network_type=network_type,
        simplify=simplify,
        retain_all=retain_all,
        truncate_by_edge=truncate_by_edge,
        clean_periphery=clean_periphery,
        custom_filter=custom_filter,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
        max_query_area_size=max_query_area_size,
    )
    utils.log(f"graph_from_address returned graph with {len(G)} nodes and {len(G.edges())} edges")

    if return_coords:
        return G, point
    else:
        return G


def graph_from_place(
    query,
    network_type="all_private",
    simplify=True,
    retain_all=False,
    truncate_by_edge=False,
    which_result=1,
    buffer_dist=None,
    clean_periphery=True,
    custom_filter=None,
    timeout=None,
    memory=None,
    custom_settings=None,
    max_query_area_size=None,
):
    """
    Create graph from OSM within the boundaries of some geocodable place(s).

    The query must be geocodable and OSM must have polygon boundaries for the
    geocode result. If OSM does not have a polygon for this place, you can
    instead get its street network using the graph_from_address function, which
    geocodes the place name to a point and gets the network within some distance
    of that point. Alternatively, you might try to vary the which_result
    parameter to use a different geocode result. For example, the first geocode
    result (ie, the default) might resolve to a point geometry, but the second
    geocode result for this query might resolve to a polygon, in which case you
    can use graph_from_place with which_result=2.

    Parameters
    ----------
    query : string or dict or list
        the place(s) to geocode/download data for
    network_type : string
        what type of street network to get if custom_filter is None. One of
        'walk', 'bike', 'drive', 'drive_service', 'all', or 'all_private'.
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected
    truncate_by_edge : bool
        if True, retain node if it's outside polygon but at least one of
        node's neighbors are within bbox
    which_result : int
        max number of results to return and which to process upon receipt
    buffer_dist : float
        distance to buffer around the place geometry, in meters
    clean_periphery : bool
        if True, buffer 500m to get a graph larger than requested, then
        simplify, then truncate it to requested spatial boundaries
    custom_filter : string
        a custom network filter to be used instead of the network_type presets,
        e.g., '["power"~"line"]' or '["highway"~"motorway|trunk"]'. Also pass
        in a network_type that is in settings.bidirectional_network_types if
        you want graph to be fully bidirectional.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module
    max_query_area_size : None
        deprecated, use ox.config(max_query_area_size=value) instead to
        configure this setting via the settings module
    Returns
    -------
    G : networkx.MultiDiGraph
    """
    # create a GeoDataFrame with the spatial boundaries of the place(s)
    if isinstance(query, (str, dict)):
        # if it is a string (place name) or dict (structured place query), then
        # it is a single place
        gdf_place = boundaries.gdf_from_place(
            query, which_result=which_result, buffer_dist=buffer_dist
        )
    elif isinstance(query, list):
        # if it is a list, it contains multiple places to get
        gdf_place = boundaries.gdf_from_places(query, buffer_dist=buffer_dist)
    else:
        raise TypeError("query must be dict, string, or list of strings")

    # extract the geometry from the GeoDataFrame to use in API query
    polygon = gdf_place["geometry"].unary_union
    utils.log("Constructed place geometry polygon(s) to query API")

    # create graph using this polygon(s) geometry
    G = graph_from_polygon(
        polygon,
        network_type=network_type,
        simplify=simplify,
        retain_all=retain_all,
        truncate_by_edge=truncate_by_edge,
        clean_periphery=clean_periphery,
        custom_filter=custom_filter,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
        max_query_area_size=max_query_area_size,
    )

    utils.log(f"graph_from_place returned graph with {len(G)} nodes and {len(G.edges())} edges")
    return G


def graph_from_polygon(
    polygon,
    network_type="all_private",
    simplify=True,
    retain_all=False,
    truncate_by_edge=False,
    clean_periphery=True,
    custom_filter=None,
    timeout=None,
    memory=None,
    custom_settings=None,
    max_query_area_size=None,
):
    """
    Create a graph from OSM within the boundaries of some shapely polygon.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        the shape to get network data within. coordinates should be in units of
        latitude-longitude degrees.
    network_type : string
        what type of street network to get if custom_filter is None. One of
        'walk', 'bike', 'drive', 'drive_service', 'all', or 'all_private'.
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected
    truncate_by_edge : bool
        if True, retain node if it's outside polygon but at least one of
        node's neighbors are within polygon
    clean_periphery : bool
        if True, buffer 500m to get a graph larger than requested, then
        simplify, then truncate it to requested spatial boundaries
    custom_filter : string
        a custom network filter to be used instead of the network_type presets,
        e.g., '["power"~"line"]' or '["highway"~"motorway|trunk"]'. Also pass
        in a network_type that is in settings.bidirectional_network_types if
        you want graph to be fully bidirectional.
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module
    max_query_area_size : None
        deprecated, use ox.config(max_query_area_size=value) instead to
        configure this setting via the settings module

    Returns
    -------
    G : networkx.MultiDiGraph
    """
    utils._handle_deprecated_params(
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
        max_query_area_size=max_query_area_size,
    )

    # verify that the geometry is valid and is a shapely Polygon/MultiPolygon
    # before proceeding
    if not polygon.is_valid:
        raise ValueError("The geometry to query within is invalid")
    if not isinstance(polygon, (Polygon, MultiPolygon)):
        raise TypeError(
            "Geometry must be a shapely Polygon or MultiPolygon. If you requested "
            "graph from place name or address, make sure your query resolves to a "
            "Polygon or MultiPolygon, and not some other geometry, like a Point. "
            "See OSMnx documentation for details."
        )

    if clean_periphery:
        # create a new buffered polygon 0.5km around the desired one
        buffer_dist = 500
        poly_proj, crs_utm = projection.project_geometry(polygon)
        poly_proj_buff = poly_proj.buffer(buffer_dist)
        poly_buff, _ = projection.project_geometry(poly_proj_buff, crs=crs_utm, to_latlong=True)

        # download the network data from OSM within buffered polygon
        response_jsons = downloader._osm_net_download(poly_buff, network_type, custom_filter)

        # create buffered graph from the downloaded data
        G_buff = _create_graph(
            response_jsons,
            retain_all=True,
            bidirectional=network_type in settings.bidirectional_network_types,
        )

        # truncate buffered graph to the buffered polygon and retain_all for
        # now. needed because overpass returns entire ways that also include
        # nodes outside the poly if the way (that is, a way with a single OSM
        # ID) has a node inside the poly at some point.
        G_buff = truncate.truncate_graph_polygon(
            G_buff, poly_buff, retain_all=True, truncate_by_edge=truncate_by_edge
        )

        # simplify the graph topology
        if simplify:
            G_buff = simplification.simplify_graph(G_buff)

        # truncate graph by original polygon to return graph within polygon
        # caller wants. don't simplify again: this allows us to retain
        # intersections along the street that may now only connect 2 street
        # segments in the network, but in reality also connect to an
        # intersection just outside the polygon
        G = truncate.truncate_graph_polygon(
            G_buff, polygon, retain_all=retain_all, truncate_by_edge=truncate_by_edge
        )

        # count how many street segments in buffered graph emanate from each
        # intersection in un-buffered graph, to retain true counts for each
        # intersection, even if some of its neighbors are outside the polygon
        G.graph["streets_per_node"] = utils_graph.count_streets_per_node(G_buff, nodes=G.nodes())

    # if clean_periphery=False, just use the polygon as provided
    else:
        # download the network data from OSM
        response_jsons = downloader._osm_net_download(polygon, network_type, custom_filter)

        # create graph from the downloaded data
        G = _create_graph(
            response_jsons,
            retain_all=True,
            bidirectional=network_type in settings.bidirectional_network_types,
        )

        # truncate the graph to the extent of the polygon
        G = truncate.truncate_graph_polygon(
            G, polygon, retain_all=retain_all, truncate_by_edge=truncate_by_edge
        )

        # simplify the graph topology as the last step. don't truncate after
        # simplifying or you may have simplified out to an endpoint beyond the
        # truncation distance, in which case you will then strip out your entire
        # edge
        if simplify:
            G = simplification.simplify_graph(G)

    utils.log(f"graph_from_polygon returned graph with {len(G)} nodes and {len(G.edges())} edges")
    return G


def graph_from_xml(filepath, bidirectional=False, simplify=True, retain_all=False):
    """
    Create a graph from data in an OSM-formatted XML file.

    Parameters
    ----------
    filepath : string
        path to file containing OSM XML data
    bidirectional : bool
        if True, create bidirectional edges for one-way streets
    simplify : bool
        if True, simplify the graph topology
    retain_all : bool
        if True, return the entire graph even if it is not connected

    Returns
    -------
    G : networkx.MultiDiGraph
    """
    # transmogrify file of OSM XML data into JSON
    response_jsons = [_overpass_json_from_file(filepath)]

    # create graph using this response JSON
    G = _create_graph(response_jsons, bidirectional=bidirectional, retain_all=retain_all)

    # simplify the graph topology as the last step.
    if simplify:
        G = simplification.simplify_graph(G)

    utils.log(f"graph_from_xml returned graph with {len(G)} nodes and {len(G.edges())} edges")
    return G


def _overpass_json_from_file(filepath):
    """
    Read OSM XML from file and return Overpass-like JSON.

    Parameters
    ----------
    filepath : string
        path to file containing OSM XML data

    Returns
    -------
    OSMContentHandler object
    """

    def _opener(filepath):
        _, ext = os.path.splitext(filepath)
        if ext == ".bz2":
            # Use Python 2/3 compatible BZ2File()
            return bz2.BZ2File(filepath)
        else:
            # Assume an unrecognized file extension is just XML
            return open(filepath, mode="rb")

    with _opener(filepath) as file:
        handler = _OSMContentHandler()
        xml.sax.parse(file, handler)
        return handler.object


def _create_graph(response_jsons, retain_all=False, bidirectional=False):
    """
    Create a networkx MultiDiGraph from Overpass API responses.

    Parameters
    ----------
    response_jsons : list
        list of dicts of JSON responses from from the Overpass API
    retain_all : bool
        if True, return the entire graph even if it is not connected
    bidirectional : bool
        if True, create bidirectional edges for one-way streets

    Returns
    -------
    G : networkx.MultiDiGraph
    """
    utils.log("Creating graph from downloaded OSM data...")

    # make sure we got data back from the server requests
    elements = []
    for response_json in response_jsons:
        elements.extend(response_json["elements"])
    if len(elements) < 1:
        raise EmptyOverpassResponse("There are no data elements in the response JSON")

    # create the graph as a MultiDiGraph and set its meta-attributes
    G = nx.MultiDiGraph(
        created_date=utils.ts(), created_with=f"OSMnx {__version__}", crs=settings.default_crs
    )

    # extract nodes and paths from the downloaded osm data
    nodes = {}
    paths = {}
    for osm_data in response_jsons:
        nodes_temp, paths_temp = _parse_osm_nodes_paths(osm_data)
        for key, value in nodes_temp.items():
            nodes[key] = value
        for key, value in paths_temp.items():
            paths[key] = value

    # add each osm node to the graph
    for node, data in nodes.items():
        G.add_node(node, **data)

    # add each osm way (ie, a path of edges) to the graph
    G = _add_paths(G, paths, bidirectional=bidirectional)

    # retain only the largest connected component, if caller did not
    # set retain_all=True
    if not retain_all:
        G = utils_graph.get_largest_component(G)

    utils.log(f"Created graph with {len(G)} nodes and {len(G.edges())} edges")

    # add length (great circle distance between nodes) attribute to each edge to
    # use as weight
    if len(G.edges) > 0:
        G = utils_graph.add_edge_lengths(G)

    return G


def _convert_node(element):
    """
    Convert an OSM node element into the format for a networkx node.

    Parameters
    ----------
    element : dict
        an OSM node element

    Returns
    -------
    node : dict
    """
    node = {}
    node["y"] = element["lat"]
    node["x"] = element["lon"]
    node["osmid"] = element["id"]
    if "tags" in element:
        for useful_tag in settings.useful_tags_node:
            if useful_tag in element["tags"]:
                node[useful_tag] = element["tags"][useful_tag]
    return node


def _convert_path(element):
    """
    Convert an OSM way element into the format for a networkx graph path.

    Parameters
    ----------
    element : dict
        an OSM way element

    Returns
    -------
    path : dict
    """
    path = {}
    path["osmid"] = element["id"]

    # remove any consecutive duplicate elements in the list of nodes
    grouped_list = groupby(element["nodes"])
    path["nodes"] = [group[0] for group in grouped_list]

    if "tags" in element:
        for useful_tag in settings.useful_tags_way:
            if useful_tag in element["tags"]:
                path[useful_tag] = element["tags"][useful_tag]
    return path


def _parse_osm_nodes_paths(osm_data):
    """
    Construct dicts of nodes and paths.

    Dict key=osmid and value=dict of attributes.

    Parameters
    ----------
    osm_data : dict
        JSON response from from the Overpass API

    Returns
    -------
    nodes, paths : tuple of dicts
    """
    nodes = {}
    paths = {}
    for element in osm_data["elements"]:
        if element["type"] == "node":
            key = element["id"]
            nodes[key] = _convert_node(element)
        elif element["type"] == "way":
            key = element["id"]
            paths[key] = _convert_path(element)

    return nodes, paths


def _add_path(G, data, one_way):
    """
    Add a path to the graph.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    data : dict
        the attributes of the path
    one_way : bool
        if this path is one-way or if it is bi-directional

    Returns
    -------
    None
    """
    # extract the ordered list of nodes from this path element, then delete it
    # so we don't add it as an attribute to the edge later
    path_nodes = data["nodes"]
    del data["nodes"]

    # set the oneway attribute to the passed-in value, to make it consistent
    # True/False values, but only do this if you aren't forcing all edges to
    # oneway with the all_oneway setting. With the all_oneway setting, you
    # likely still want to preserve the original OSM oneway attribute.
    if not settings.all_oneway:
        data["oneway"] = one_way

    # zip together the path nodes so you get tuples like (0,1), (1,2), (2,3)
    # and so on
    path_edges = list(zip(path_nodes[:-1], path_nodes[1:]))
    G.add_edges_from(path_edges, **data)

    # if the path is NOT one-way
    if not one_way:
        # reverse the direction of each edge and add this path going the
        # opposite direction
        path_edges_opposite_direction = [(v, u) for u, v in path_edges]
        G.add_edges_from(path_edges_opposite_direction, **data)


def _add_paths(G, paths, bidirectional=False):
    """
    Add a collection of paths to the graph.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    paths : dict
        the paths from OSM
    bidirectional : bool
        if True, create bidirectional edges for one-way streets

    Returns
    -------
    None
    """
    # the list of values OSM uses in its 'oneway' tag to denote True
    # https://www.geofabrik.de/de/data/geofabrik-osm-gis-standard-0.7.pdf
    osm_oneway_values = ["yes", "true", "1", "-1", "T", "F"]

    for data in paths.values():

        if settings.all_oneway is True:
            _add_path(G, data, one_way=True)
        # if this path is tagged as one-way and if it is not a walking network,
        # then we'll add the path in one direction only
        elif ("oneway" in data and data["oneway"] in osm_oneway_values) and not bidirectional:
            if data["oneway"] == "-1" or data["oneway"] == "T":
                # paths with a one-way value of -1 or T are one-way, but in the
                # reverse direction of the nodes' order, see osm documentation
                data["nodes"] = list(reversed(data["nodes"]))
            # add this path (in only one direction) to the graph
            _add_path(G, data, one_way=True)

        elif ("junction" in data and data["junction"] == "roundabout") and not bidirectional:
            # roundabout are also oneway but not tagged as is
            _add_path(G, data, one_way=True)

        # else, this path is not tagged as one-way or it is a walking network
        # (you can walk both directions on a one-way street)
        else:
            # add this path (in both directions) to the graph and set its
            # 'oneway' attribute to False. if this is a walking network, this
            # may very well be a one-way street (as cars/bikes go), but in a
            # walking-only network it is a bi-directional edge
            _add_path(G, data, one_way=False)

    return G


class _OSMContentHandler(xml.sax.handler.ContentHandler):
    """
    SAX content handler for OSM XML.

    Used to build an Overpass-like response JSON object in self.object. For format
    notes, see http://wiki.openstreetmap.org/wiki/OSM_XML#OSM_XML_file_format_notes
    and http://overpass-api.de/output_formats.html#json
    """

    def __init__(self):
        self._element = None
        self.object = {"elements": []}

    def startElement(self, name, attrs):
        if name == "osm":
            self.object.update({k: attrs[k] for k in attrs.keys() if k in ("version", "generator")})

        elif name in ("node", "way"):
            self._element = dict(type=name, tags={}, nodes=[], **attrs)
            self._element.update({k: float(attrs[k]) for k in attrs.keys() if k in ("lat", "lon")})
            self._element.update(
                {
                    k: int(attrs[k])
                    for k in attrs.keys()
                    if k in ("id", "uid", "version", "changeset")
                }
            )

        elif name == "tag":
            self._element["tags"].update({attrs["k"]: attrs["v"]})

        elif name == "nd":
            self._element["nodes"].append(int(attrs["ref"]))

        elif name == "relation":
            # Placeholder for future relation support.
            # Look for nested members and tags.
            pass

    def endElement(self, name):
        if name in ("node", "way"):
            self.object["elements"].append(self._element)



"""Download and plot footprints from OpenStreetMap."""

import geopandas as gpd
from shapely.geometry import LineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
from shapely.ops import polygonize

from . import boundaries
from . import downloader
from . import projection
from . import settings
from . import utils
from . import utils_geo


def _osm_footprints_download(polygon, footprint_type="building"):
    """
    Download OpenStreetMap footprint data as a list of json responses.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        geographic boundaries to fetch the footprints within
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building', 'landuse', 'place', etc.

    Returns
    -------
    response_jsons : list
        list of response_json dicts
    """
    response_jsons = []
    overpass_settings = downloader._make_overpass_settings()

    # project to utm, divide polygon up into sub-polygons if area exceeds a
    # max size (in meters), project back to lat-lng, then get a list of polygon(s) exterior coordinates
    geometry_proj, crs_proj = projection.project_geometry(polygon)
    gpcs = utils_geo._consolidate_subdivide_geometry(geometry_proj)
    geometry, _ = projection.project_geometry(gpcs, crs=crs_proj, to_latlong=True)
    polygon_coord_strs = utils_geo._get_polygons_coordinates(geometry)
    utils.log(
        f"Requesting footprints within polygon from API in {len(polygon_coord_strs)} request(s)"
    )

    # pass each polygon exterior coordinates in the list to the API, one at
    # a time
    for polygon_coord_str in polygon_coord_strs:
        query_str = (
            f"{overpass_settings};("
            f'way(poly:"{polygon_coord_str}")["{footprint_type}"];(._;>;);'
            f'relation(poly:"{polygon_coord_str}")["{footprint_type}"];(._;>;););out;'
        )
        response_json = downloader.overpass_request(data={"data": query_str})
        response_jsons.append(response_json)
    utils.log(
        f"Got all footprint data within polygon from API in {len(polygon_coord_strs)} request(s)"
    )

    return response_jsons


def _create_footprints_gdf(
    polygon=None, footprint_type="building", retain_invalid=False, responses=None
):
    """
    Get footprint (polygon) data from OSM and convert it into a GeoDataFrame.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        geographic boundaries to fetch the footprints within
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building',
        'landuse', 'place', etc.
    retain_invalid : bool
        if False discard any footprints with an invalid geometry
    responses : list
        list of pre-existing response jsons

    Returns
    -------
    gdf : geopandas.GeoDataFrame
    """
    if polygon is None and responses is None:
        raise ValueError("You must pass a polygon or pre-existing responses.")

    if responses is None:
        responses = _osm_footprints_download(polygon, footprint_type)

    # parse the list of responses into separate dicts of vertices, footprints and relations
    # create a set of ways not directly tagged with footprint_type
    vertices, footprints, relations, untagged_ways = _responses_to_dicts(responses, footprint_type)

    # create simple Shapely geometries (Polygon or LineString) for all of the ways in footprints
    for footprint_key, footprint_val in footprints.items():
        footprint_val["geometry"] = _create_footprint_geometry(
            footprint_key, footprint_val, vertices
        )

    # create a complex Shapely Polygon or MultiPolygon for each relation
    for relation_key, relation_val in relations.items():
        relation_val["geometry"] = _create_relation_geometry(relation_key, relation_val, footprints)

    # merge relations into the footprints dictionary
    footprints.update(relations)

    # delete supporting geometry not directly tagged with footprint_type from the footprints dictionary
    for untagged_way in untagged_ways:
        try:
            del footprints[untagged_way]
        except KeyError:
            utils.log(f"untagged_way {untagged_way} not found in footprints dict")

    # Convert footprints dictionary to a GeoDataFrame
    gdf = gpd.GeoDataFrame.from_dict(footprints, orient="index")
    gdf.crs = settings.default_crs

    # filter the gdf to only include valid Polygons/MultiPolygons if retain_invalid is False
    if not retain_invalid and not gdf.empty:
        filter1 = gdf["geometry"].is_valid
        filter2 = (gdf["geometry"].geom_type == "Polygon") | (
            gdf["geometry"].geom_type == "MultiPolygon"
        )
        filter_combined = filter1 & filter2
        gdf = gdf[filter_combined]

    return gdf


def _responses_to_dicts(responses, footprint_type):
    """
    Parse list of json responses into dicts of vertices, footprints, relations.

    Note: OSM's data model and the Overpass API will return open ways (lines)
    as part of a 'polygon' query. These may be fragments of the inner and
    outer rings of relations or they may be open ways mistakenly tagged with
    'polygon' type tags.

    Ways not directly tagged with the footprint type are added to the
    untagged_ways set for removal from the footprints dictionary at the end of
    the process.

    Some inner ways of relations may be tagged with the footprint type in
    their own right e.g. landuse=meadow as an inner way in a landuse=forest
    relation and need to be kept. These are created here.

    Parameters
    ----------
    responses : list
        list of json responses
    footprint_type : string
        type of footprint downloaded. OSM tag key e.g. 'building', 'landuse',
        'place', etc.

    Returns
    -------
    (vertices, footprints, relations, untagged_footprints) : tuple
        vertices
            dictionary of OSM nodes including their lat, lng coordinates
        footprints
            dictionary of OSM ways including their nodes and tags
        relations
            dictionary of OSM relations including member ids and tags
        untagged_footprints
            set of ids for ways or relations not directly tagged with
            footprint_type
    """
    # create dictionaries to hold vertices, footprints and relations
    vertices = {}
    footprints = {}
    relations = {}

    # create a set to hold the ids of ways not directly tagged as footprint_type
    untagged_footprints = set()

    # loop through each response once adding each element to one of the dicts
    for response in responses:
        for element in response["elements"]:
            # NODES - only keep coordinates
            if "type" in element and element["type"] == "node":
                vertices[element["id"]] = {"lat": element["lat"], "lon": element["lon"]}
            # WAYS - both open and closed
            elif "type" in element and element["type"] == "way":
                footprint = {"nodes": element["nodes"]}
                if "tags" in element:
                    for tag in element["tags"]:
                        footprint[tag] = element["tags"][tag]
                footprints[element["id"]] = footprint
                # add ways not individually tagged with footprint_type to the
                # untagged_footprints set
                if ("tags" not in element) or (footprint_type not in element["tags"]):
                    untagged_footprints.add(element["id"])
            # RELATIONS
            elif "type" in element and element["type"] == "relation":
                relation = {"members": {}}
                for member in element["members"]:
                    if "type" in member and member["type"] == "way":
                        relation["members"].update({member["ref"]: member.get("role")})
                if "tags" in element:
                    for tag in element["tags"]:
                        relation[tag] = element["tags"][tag]
                relations[element["id"]] = relation
                # add relations not individually tagged with footprint_type to
                # the untagged_footprints set
                if ("tags" not in element) or (footprint_type not in element["tags"]):
                    untagged_footprints.add(element["id"])
            else:
                utils.log(f'Element {element["id"]} is not a node, way or relation')

    return vertices, footprints, relations, untagged_footprints


def _create_footprint_geometry(footprint_key, footprint_val, vertices):
    """
    Create geometry for footprint open/closed ways.

    Create Shapely geometry for open or closed ways in the initial footprints
    dictionary. Closed ways are converted directly to Shapely Polygons, open
    ways (fragments that will form the outer and inner rings of relations) are
    converted to LineStrings.

    Parameters
    ----------
    footprint_key : int
        the id of the way/footprint to process
    footprint_val : dict
        the nodes and tags of the footprint
    vertices : dict
        the dictionary of OSM nodes with their coordinates

    Returns
    -------
    Shapely Polygon or LineString
    """
    # CLOSED WAYS
    if footprint_val["nodes"][0] == footprint_val["nodes"][-1]:
        try:
            poly = [
                (vertices[node]["lon"], vertices[node]["lat"]) for node in footprint_val["nodes"]
            ]
            footprint_geometry = Polygon(poly)
        except Exception:
            utils.log(f"Polygon has invalid geometry: {footprint_key}")
    # OPEN WAYS
    else:
        try:
            ls = [(vertices[node]["lon"], vertices[node]["lat"]) for node in footprint_val["nodes"]]
            footprint_geometry = LineString(ls)
        except Exception:
            utils.log(f"LineString has invalid geometry: {footprint_key}")

    return footprint_geometry


def _create_relation_geometry(relation_key, relation_val, footprints):
    """
    Create Shapely geometry for relations: Polygons with holes or MultiPolygons.

    OSM relations are used to define complex polygons - polygons with holes or
    multi-polygons. The polygons' outer and inner rings may be made up of chains
    of LineStrings. https://wiki.openstreetmap.org/wiki/Relation:multipolygon
    requires that multipolygon rings have an outer or inner 'role'.

    OSM's data model allows a polygon type tag e.g. 'building' to be added to
    any OSM element. This can include non-polygon relations e.g. bus routes.
    Relations that do not have at least one closed ring with an outer role
    are filtered out.

    Inner rings that are tagged with the footprint type in their own right e.g.
    landuse=meadow as an inner ring of landuse=forest will have been included in
    the footprints dictionary as part of the original parsing and are not dealt
    with here.

    Parameters
    ----------
    relation_key : int
        the id of the relation to process
    relation_val : dict
        members and tags of the relation
    footprints : dict
        dictionary of all footprints (including open and closed ways)

    Returns
    -------
    Shapely Polygon or MultiPolygon
    """
    # lists to hold member geometries
    outer_polys, outer_lines, inner_polys, inner_lines = _members_geom_lists(
        relation_val, footprints
    )

    # try to polygonize open outer ways and concatenate them to outer_polys
    if len(outer_lines) > 0:
        try:
            result = list(polygonize(outer_lines))
        except Exception:
            utils.log(f"polygonize failed for outer ways in relation: {relation_key}")
        else:
            outer_polys += result

    # try to polygonize open inner ways and concatenate them to inner_polys
    if len(inner_lines) > 0:
        try:
            result = list(polygonize(inner_lines))
        except Exception:
            utils.log(f"polygonize failed for inner ways in relation: {relation_key}")
        else:
            inner_polys += result

    # filter out relations missing both 'outer' and 'inner' polygons or just 'outer'
    multipoly = []
    if len(outer_polys + inner_polys) == 0:
        utils.log(f"Relation {relation_key} missing outer and inner closed ways")
    elif len(outer_polys) == 0:
        utils.log(f"Relation {relation_key} missing outer closed ways")
    # process the others to multipolygons
    else:
        for outer_poly in outer_polys:
            outer_poly = outer_poly.buffer(0)  # fix invalid geometry if present
            temp_poly = outer_poly
            for inner_poly in inner_polys:
                inner_poly = inner_poly.buffer(0)  # fix invalid geometry if present
                if inner_poly.within(outer_poly):
                    temp_poly = temp_poly.difference(inner_poly)
            multipoly.append(temp_poly)

    # return relations with one outer way as Polygons, multiple outer ways
    # as MultiPolygons
    if len(multipoly) == 1:
        return multipoly[0]
    elif len(multipoly) > 1:
        return MultiPolygon(multipoly)
    else:
        utils.log(f"relation {relation_key} could not be converted to a complex footprint")


def _members_geom_lists(relation_val, footprints):
    """
    Add relation members' geoms to lists.

    Parameters
    ----------
    relation_val : dict
        members and tags of the relation
    footprints : dict
        dictionary of all footprints (including open and closed ways)

    Returns
    -------
    tuple of lists
    """
    outer_polys = []
    outer_lines = []
    inner_polys = []
    inner_lines = []

    # add each members geometry to a list according to its role and geometry type
    for member_id, member_role in relation_val["members"].items():
        if member_role == "outer":
            if footprints[member_id]["geometry"].geom_type == "Polygon":
                outer_polys.append(footprints[member_id]["geometry"])
            elif footprints[member_id]["geometry"].geom_type == "LineString":
                outer_lines.append(footprints[member_id]["geometry"])
        elif member_role == "inner":
            if footprints[member_id]["geometry"].geom_type == "Polygon":
                inner_polys.append(footprints[member_id]["geometry"])
            elif footprints[member_id]["geometry"].geom_type == "LineString":
                inner_lines.append(footprints[member_id]["geometry"])

    return outer_polys, outer_lines, inner_polys, inner_lines


def footprints_from_point(
    point,
    dist=1000,
    footprint_type="building",
    retain_invalid=False,
    timeout=None,
    memory=None,
    custom_settings=None,
):
    """
    Get footprints within some distance N, S, E, W of a lat-lng point.

    Parameters
    ----------
    point : tuple
        a lat-lng point
    dist : numeric
        distance in meters
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building',
        'landuse', 'place', etc.
    retain_invalid : bool
        if False discard any footprints with an invalid geometry
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    geopandas.GeoDataFrame
    """
    north, south, east, west = utils_geo.bbox_from_point(point=point, dist=dist)
    polygon = utils_geo.bbox_to_poly(north, south, east, west)
    return footprints_from_polygon(
        polygon,
        footprint_type,
        retain_invalid,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
    )


def footprints_from_address(
    address,
    dist=1000,
    footprint_type="building",
    retain_invalid=False,
    timeout=None,
    memory=None,
    custom_settings=None,
):
    """
    Get footprints within some distance N, S, E, W of an address.

    Parameters
    ----------
    address : string
        the address to geocode to a lat-lng point
    dist : numeric
        distance in meters
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building',
        'landuse', 'place', etc.
    retain_invalid : bool
        if False discard any footprints with an invalid geometry
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    geopandas.GeoDataFrame
    """
    # geocode the address string to a (lat, lng) point
    point = utils_geo.geocode(query=address)

    # get footprints within distance of this point
    return footprints_from_point(
        point,
        dist,
        footprint_type,
        retain_invalid,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
    )


def footprints_from_place(
    place,
    footprint_type="building",
    retain_invalid=False,
    which_result=1,
    timeout=None,
    memory=None,
    custom_settings=None,
):
    """
    Get footprints within the boundaries of some place.

    The query must be geocodable and OSM must have polygon boundaries for the
    geocode result. If OSM does not have a polygon for this place, you can
    instead get its footprints using the footprints_from_address function, which
    geocodes the place name to a point and gets the footprints within some distance
    of that point.

    Parameters
    ----------
    place : string
        the query to geocode to get geojson boundary polygon
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building',
        'landuse', 'place', etc.
    retain_invalid : bool
        if False discard any footprints with an invalid geometry
    which_result : int
        max number of results to return and which to process upon receipt
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    geopandas.GeoDataFrame
    """
    city = boundaries.gdf_from_place(place, which_result=which_result)
    polygon = city["geometry"].iloc[0]
    return footprints_from_polygon(
        polygon,
        footprint_type,
        retain_invalid,
        timeout=timeout,
        memory=memory,
        custom_settings=custom_settings,
    )


def footprints_from_polygon(
    polygon,
    footprint_type="building",
    retain_invalid=False,
    timeout=None,
    memory=None,
    custom_settings=None,
):
    """
    Get footprints within some polygon.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
        the shape to get data within. coordinates should be in units of
        latitude-longitude degrees.
    footprint_type : string
        type of footprint to be downloaded. OSM tag key e.g. 'building',
        'landuse', 'place', etc.
    retain_invalid : bool
        if False discard any footprints with an invalid geometry
    timeout : None
        deprecated, use ox.config(timeout=value) instead to configure this
        setting via the settings module
    memory : None
        deprecated, use ox.config(memory=value) instead to configure this
        setting via the settings module
    custom_settings : None
        deprecated, use ox.config(custom_settings=value) instead to configure
        this setting via the settings module

    Returns
    -------
    geopandas.GeoDataFrame
    """
    utils._handle_deprecated_params(timeout=timeout, memory=memory, custom_settings=custom_settings)
    return _create_footprints_gdf(polygon, footprint_type, retain_invalid)
